package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/**
 * A <code>PolyJConstructorCall</code> is an immutable representation of
 * a direct call to a constructor of a class in the form of <code>super(...)</code>
 * or <code>this(...)</code>.  It consists of the kind of the call and a list of
 * expressions to be parameters of the call.  A constructor call statement may also
 * contain an expression providing the context in which it is executed.  This AST node
 * is extended because the constructors in parametrized classes change to include a where
 * class, and thus calls to the constructor must be appropriately translated.
 **/
public class PolyJConstructorCall_c extends ConstructorCall_c implements PolyJConstructorCall {

    /**
     * Creates a new <code>PolyJConstructorCall</code>.
     *
     * @pre <code>arguments</code> must be a list of <code>Expr</code>.
     */
    public PolyJConstructorCall_c(Position pos, Kind kind,
            Expr qualifier, List arguments)
    {
        super(pos, kind, qualifier, arguments);
    }

    /**
     * Since constructors for parametrized classes are not translated transparently,
     * we must translate constructor calls to account for the change to constructors
     */
    public void translate(CodeWriter w, Translator tr) {
        Context c = tr.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();

        if (constructorInstance().container() instanceof InstType) {
            translateInstConst(w, tr, ts, c);
        } else {
            translateDefault(w, tr);
        }
        w.write(";");
    }

    /**
     * The default translation is to translate the qualifier
     * and kind normally, and to translate the args normally
     * except to add a where object if it's needed as the
     * first argument.
     */
    protected void translateDefault(CodeWriter w, Translator tr) {
        translateQualifier(w, tr);
        translateKind(w, tr);
        PolyJTranslator.translateArgs(w, tr, arguments(), constructorInstance());
    }

    /**
     * Translate a call to a constructor of an instantiated type.
     * This may require boxing/unboxing of primitives if the
     * type was instantiated on primitives.
     */
    protected void translateInstConst(CodeWriter w, Translator tr,
            PolyJTypeSystem ts, Context c) {
        ConstructorInstance myCi = constructorInstance();
        ConstructorInstance origCi = ts.originalConstructor(myCi);

        translateQualifier(w, tr);
        translateKind(w, tr);

        if (kind == SUPER) {
            if (origCi == null) {
                throw new InternalCompilerError("Found a constructor in an "
                    + "instantiation of a polymorphic type that is not "
                    + " in the original type!");
            }
            PolyJTranslator.translateInstArgs(w, tr, ts, c, arguments(), null,
                myCi, origCi);
        } else { // kind == THIS
            // We don't want to box arguments if we're
            // calling an alternate constructor in the same class.
            PolyJTranslator.translateArgs(w, tr, arguments(), myCi);
        }
    }

    /**
     * Translate the qualifier of the call.
     */
    protected void translateQualifier(CodeWriter w, Translator tr) {
        if (qualifier != null) {
            qualifier.del().translate(w, tr);
            w.write(".");
        }
    }

    /**
     * Translate the kind of the call (this or super).
     */
    protected void translateKind(CodeWriter w, Translator tr) {
        w.write("" + kind);
    }

}
