package polyglot.ext.polyj.types;

import polyglot.ext.polyj.visit.*;

import polyglot.types.*;
import polyglot.util.*;

import java.util.*;

/**
 * Each InstType keeps a list of substitutions
 * for each of the parameters to its polymorphic type.
 * additionally, it also has each of the substitutions for
 * the enclosing polymorphic classes.  Eg:
 *
 *  class A[X] { class B { class C[Y,Z] { class D { ... }}}}
 *
 *  then for the expression A[int].B.C[A[Object], W].D
 *  the enclosing class of D is a InstType with the
 *  substitutions: { X|->int, Y|->A[Object], Z|->W }
 *  (in the example I use a different letter for each parameter
 *   type, but in fact we use our usual notion of parameter equality
 *   rather than byname equality)
 *
 *  The InstType does lazy substitution.  We do no
 *  substitutions at the outset, and only perform them when someone
 *  queries us for a method, field, interface, supertype, inner class,
 *  etc.
 */
public class InstType_c extends polyglot.ext.jl.types.ClassType_c implements InstType
{
    PolyJTypeWrapper tw;
    /**
     * A subset of the value set of the substitutions map that
     * corresponds to the actual arguments in the instantiation
     * of this type.
     */
    List actualArgs;
    PolyJClassType instantiatedFrom;

    /**
     * A map from the methods of the polymorphic type to the methods
     * of this instantiation of that type.
     * MTI of this => MTI of instantiatedFrom
     */
    Map uninstantiatedMethods;
    Map uninstantiatedConstructors;

    InstType_c(Position pos, PolyJClassType instantiatedFrom,
                    List actualArgs, PolyJTypeWrapper tw)
    {
        super(instantiatedFrom.typeSystem(), pos);
        this.instantiatedFrom = instantiatedFrom;
        this.tw = tw;
        this.actualArgs = TypedList.copyAndCheck(actualArgs, Type.class, true);
        this.uninstantiatedMethods = new HashMap();
        this.uninstantiatedConstructors = new HashMap();
    }

    InstType_c(Position pos, PolyJClassType instantiatedFrom, 
                    List actualArgs, PolyJTypeSystem ts)
    {
        this(pos, instantiatedFrom, actualArgs, 
            ts.Wrapper(instantiatedFrom, actualArgs));
    }

    public Kind kind() {
        return instantiatedFrom.kind();
    }

    public boolean inStaticContext() {
        return instantiatedFrom.inStaticContext();
    }

    public List actualArgs() {
        return actualArgs;
    }

    public PolyJTypeWrapper substitutor() {
        return tw;
    }
    
    /**
     * true if the polymorphic type was instantiated on its own formal params
     */
    public boolean isNullInstantiation() {
        return actualArgs().equals(instantiatedFrom.params());
    }

    public PolyJClassType instType() { return instantiatedFrom; }

    public MethodInstance originalMethod(MethodInstance mi) {
        return (MethodInstance) uninstantiatedMethods.get(mi);
    }
    
    public ConstructorInstance originalConstructor(ConstructorInstance ci) {
        return (ConstructorInstance) uninstantiatedConstructors.get(ci);
    }

    public int hashCode() {
        return instantiatedFrom.hashCode();
        // We can't include actualArgs in the hash because
        // if this is a null instantiatation, it
        // can be equal to a PolyJClassType.  A PJCT doesn't
        // have actuals, so including them here would
        // cause their hashes to differ.  That would break
        // the invariant that if two objects are equals()
        // they must have the same hashCode().
    }

    public boolean equalsImpl(TypeObject t) {
        PolyJTypeSystem ts = (PolyJTypeSystem) typeSystem();
	if (t instanceof ClassType) {
	    ClassType c = (ClassType) t;
            if (ts.isInstantiated(c)) {
                return ts.isSameInst(this, c);
	    }
        }
        return false;
    }

    public polyglot.types.Package package_() {
        return instantiatedFrom.package_();
    }

    public String fullName() {
        return PolyJTranslator.fullNameForClass(instantiatedFrom);
    }

    public String name() {
        return PolyJTranslator.nameForClass(instantiatedFrom);
    }

    public String toString() {
        if (ts.equals(this, instType())) {
            return instType().toString();
        } else {
            return PolyJTranslator.instTypeString(instType(), actualArgs());
        }
    }

    protected String addArgs() {
        String s = "[";
        Iterator it = actualArgs().iterator();
        if (it.hasNext()) {
            s += ((Type)it.next()).toString();
        }
        while (it.hasNext()) {
            s += ", " + ((Type)it.next()).toString();
        }
        s += "]";
        return s;  }

    public String translate(Resolver c) {
        return PolyJTranslator.fullNameForClass(instantiatedFrom);
    }

    public List methods() {
        return tw.methods(instantiatedFrom, instantiatedFrom, this,
            uninstantiatedMethods);
    }

    public List constructors() {
        return tw.constructors(instantiatedFrom, instantiatedFrom, this,
            uninstantiatedConstructors);
    }

    public List fields() {
        return tw.fields(instantiatedFrom,
                            instantiatedFrom,this);
    }

    public Flags flags() {
        return instantiatedFrom.flags();
    }
    
    public Type superType() { 
        return tw.superType(instantiatedFrom, this);
    }
    
    public List interfaces() {
        return tw.interfaces(instantiatedFrom, this);
    }
    
    public boolean isNested() { return instantiatedFrom.isNested(); }

    public boolean isAnonymous() {
        return instantiatedFrom.isAnonymous();
    }

    public boolean isLocal() { return instantiatedFrom.isLocal(); }

    public ClassType outer() {
        return instantiatedFrom.outer();
    }

    public List memberClasses() {
        return tw.memberClasses(instantiatedFrom,
                                  instantiatedFrom, this);
    }

    public ClassType memberClassNamed(String name) {
        return tw.memberClassNamed(name,instantiatedFrom,
                                instantiatedFrom, this);
    }

    public boolean isEnclosedImpl(ClassType maybe_outer) {
        if (! isInner()) {
            return false;
        }

        PolyJTypeSystem ts = (PolyJTypeSystem) this.ts;

        if (! ts.isPolymorphicClass(maybe_outer)) {
            return false;
        }

        if (! instantiatedFrom.isEnclosed(ts.uninstantiate(maybe_outer))) {
            return false;
        }

        return ts.equals(maybe_outer, tw.wrap(ts.uninstantiate(maybe_outer),
                                          instantiatedFrom, this));
    }
}
